/*
 * This script displays some information about the distance
 * between two track ends.
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2018-12-14
 */

//http://www.ribbonsoft.com/doc/qcad/3.0/developer/class_r_entity.html

include("../QCTrack.js");

function AngleHint(guiAction) {
    EAction.call(this, guiAction);
}

AngleHint.prototype = new EAction();
AngleHint.includeBasePath = includeBasePath;

AngleHint.State = {
    SettingLineStart : 0,
    SettingLineEnd : 1
};

AngleHint.prototype.beginEvent = function() {
    EAction.prototype.beginEvent.call(this);
    this.eps = RMath.deg2rad(0.001);
    this.setState(AngleHint.State.SettingLineStart);
};

AngleHint.prototype.setState = function(state) {
    EAction.prototype.setState.call(this, state);

    this.getDocumentInterface().setClickMode(RAction.PickEntity);
    this.setCrosshairCursor();

    var appWin = RMainWindowQt.getMainWindow();
    var info;
    switch (this.state) {
    case AngleHint.State.SettingLineStart:
      this.lineSource = undefined;
      this.lineDest = undefined;
      info = qsTr("Select first track end");
      appWin.setCommandPrompt(info);
      appWin.setLeftMouseTip(info);
      appWin.setRightMouseTip(EAction.trCancel);
      break;
    case AngleHint.State.SettingLineEnd:
      this.lineDest = undefined;
      info = qsTr("Select second track end");
      appWin.setCommandPrompt(info);
      appWin.setLeftMouseTip(info);
      appWin.setRightMouseTip(EAction.trBack);
      break;
    }
};

AngleHint.prototype.escapeEvent = function() {
    switch (this.state) {
    case AngleHint.State.SettingLineStart:
      EAction.prototype.escapeEvent.call(this);
      break;
    case AngleHint.State.SettingLineEnd:
      this.setState(AngleHint.State.SettingLineStart);
      break;
    }
};

AngleHint.prototype.pickEntity = function(event, preview) {
    var di = this.getDocumentInterface();
    var document = this.getDocument();
    var entityId = event.getEntityId();
    var entity = document.queryEntity(entityId);
    var pos = event.getModelPosition();
    var op;

    switch (this.state) {
    case AngleHint.State.SettingLineStart:
      if (isNull(entity)) {
        this.lineStart = undefined;
        this.middleStart = undefined;
        this.lineEnd = undefined;
        this.middleEnd = undefined;
        if (preview) {
            this.updatePreview();
        }
        break;
      }
      this.lineStart = getClosestTrackEnd(document, entity, pos, false);
      if (this.lineStart == undefined) {
        this.middleStart = undefined;
        if (preview) {
            this.updatePreview();
        }
        break;
      }
      this.middleStart = this.lineStart.getMiddlePoints()[0];
      if (preview) {
        this.updatePreview();
      }
      else {
        this.setState(AngleHint.State.SettingLineEnd);
      }
      break;
    case AngleHint.State.SettingLineEnd:
      if (isNull(entity)) {
        this.lineEnd = undefined;
        if (preview) {
            this.updatePreview();
        }
        break;
      }
      this.lineEnd = getClosestTrackEnd(document, entity, pos, false);
      if (this.lineEnd == undefined) {
        this.middleEnd = undefined;
        if (preview) {
            this.updatePreview();
        }
        break;
      }
      this.middleEnd = this.lineEnd.getMiddlePoints()[0];
      if (preview) {
        this.updatePreview();
      }
      else {
        EAction.handleUserMessage(qsTr("info from ")
          + this.middleStart.getX().toFixed(1) + " / " + this.middleStart.getY().toFixed(1) + qsTr(" at ")
          + RMath.rad2deg(this.lineStart.getAngle()).toFixed(2) + qsTr("° to ")
          + this.middleEnd.getX().toFixed(1) + " / " + this.middleEnd.getY().toFixed(1) + qsTr(" at ")
          + RMath.rad2deg(this.lineEnd.getAngle()).toFixed(2) + qsTr("°"));
        var angleDiff = getNormalizedAngle(this.lineEnd.getAngle() - this.lineStart.getAngle(), Math.PI, this.eps);
        if (angleDiff > Math.PI / 2) angleDiff = Math.PI - angleDiff;
        var middleDiff = this.middleEnd.getDistanceTo(this.middleStart);
        EAction.handleUserMessage(qsTr("direct distance: ") + middleDiff.toFixed(1) + qsTr(", track end rotated by ")
          + RMath.rad2deg(angleDiff).toFixed(2) + qsTr("°"));
        //ok, got the target points
        //now get a list of angles in all (curved) tracks
        this.list = new Array(); //each element holds the possible angle with 0 < angle < 180° and the maximum count to use
        var allIds = document.queryAllBlocks();
        for (var i = 0; i < allIds.length; i++) { //first run: collect blocks with exactly one straight line in TR_MIDS
          var id = allIds[i];
          var block = document.queryBlockDirect(id);
          var blockName = block.getName();
          if (blockName[0] == '*') { //reserved blocks like paper space
            continue;
          }
          var subIds;
          subIds = document.queryBlockEntities(id);
          var blockAngle = 0;
          var blockEnds = 0;
          for (var j = 0; j < subIds.length; j++) {
            var sub = document.queryEntity(subIds[j]);
            if (sub.getLayerName() == "TR_MIDS") {
              if (blockAngle > 0) { //block has another entity on TR_MIDS layer
                blockAngle = -1;
                break;
              }
              if (isArcEntity(sub)) {
                blockAngle = getNormalizedAngle(sub.getEndAngle() - sub.getStartAngle(), Math.PI * 2, this.eps);
              }
              else { //not (just) a curved arc
                break;
              }
            }
            else if (sub.getLayerName() == "TR_ENDS") {
              blockEnds++;
            }
          }
          if (blockAngle > 0 && blockEnds == 2) {
            var found = false;
            for (var j = 0; j < this.list.length; j++) {
              if (this.list[j][0] == blockAngle) {
                found = true;
                break;
              }
            }
            if (!found) {
              this.list.push([blockAngle,0]);
            }
          }
        }
        if (this.list.length == 0) {
          EAction.handleUserMessage(qsTr("no curved track found"));
          (new QMessageBox(null, qsTr("AngleHint"), qsTr("no curved track found"),
            new QMessageBox.StandardButtons(QMessageBox.Ok), null, Qt.Dialog)).exec();
          return false;
        }
        this.list.sort(function(a, b) {
          return b[0] - a[0];
        });

        var degList = "";
        for (var i = 0; i < this.list.length; i++) {
          degList += " " + RMath.rad2deg(this.list[i][0]).toFixed(2);
        }
        this.tarAngle = angleDiff;
        this.maxElems = 15;
        this.maxDiff = RMath.deg2rad(1.0);
        EAction.handleUserMessage(qsTr("searching angle ") + RMath.rad2deg(this.tarAngle).toFixed(2)
          + qsTr("° +/- ") + RMath.rad2deg(this.maxDiff).toFixed(2)
          + qsTr("° as combination of max. ") + this.maxElems + qsTr(" tracks of ") + degList);

        this.solutions = new Array(); //will contain all arrays with [0]: array containing angle and element count and [1]: solution arrays particular counts
        this.extendCurrentCombination(0); //do all the work
        this.removeDuplicates(); //remove combinations with (nearly) the same angle
        this.solutions.sort(function(a, b) { //sort the list for output
          var d = a[0][0] - b[0][0]; //get angle difference
          if (Math.abs(d) >= this.eps) {
            return d;
          }
          return a[0][1] - b[0][1]; //if angle is equal, compare element count
        });

        for (var i = 0; i < this.solutions.length; i++) {
          var delta = this.solutions[i][0][0] - this.tarAngle;
          if (Math.abs(delta) < this.eps) delta = qsTr("+-");
          else delta = RMath.rad2deg(delta).toFixed(2);
          var msg = (i + 1) + ": " + RMath.rad2deg(this.solutions[i][0][0]).toFixed(2) + qsTr("° (")
            + delta + qsTr("°) done with ") + this.solutions[i][0][1] + qsTr(" tracks");
          for (var j = 0; j < this.list.length; j++) {
            var cnt = this.getFactor(this.solutions[i][1][j]);
            if (cnt != 0) {
              msg += qsTr(", ") + cnt + qsTr("x ") + RMath.rad2deg(this.list[j][0]).toFixed(2) + qsTr("°");
            }
          }
          EAction.handleUserMessage(msg);
        }

        EAction.handleUserMessage(qsTr("found ") + this.solutions.length + qsTr(" solutions"));
        this.setState(-1);
        this.terminate();
      }
      break;
    }
};

AngleHint.prototype.extendCurrentCombination = function(startAt) {
    var curAngle = this.getCurrentAngle();
    var curElems = this.getCurrentElementCount();
    if (curAngle > this.tarAngle + this.maxDiff || curElems > this.maxElems) { //exceeded maximum angle or maximum element count
      return;
    }
    if (startAt < this.list.length) {
      while (curElems <= this.maxElems && curAngle <= this.tarAngle + this.eps) { //maximum element count and target angle must not be exceeded
        this.extendCurrentCombination(startAt + 1);
        this.list[startAt][1]++;
        curAngle = this.getCurrentAngle();
        curElems = this.getCurrentElementCount();
      }
      this.list[startAt][1]=0; //reset for further back-tracking
    }
    else { //all elements done (maybe some are ==0), check if solution found
      if (Math.abs(curAngle - this.tarAngle) < this.maxDiff) { //found a solution
        var sol = this.getSolutionArray();
        var angle = this.getCombinationAngle(sol);
        var elem = this.getCombinationElementCount(sol);
        this.solutions.push([[angle, elem], sol]);
      }
    }
};

AngleHint.prototype.getCurrentAngle = function() {
    var angle = 0;
    for (var i = 0; i < this.list.length; i++) {
      angle += this.list[i][0] * this.getFactor(this.list[i][1]);
    }
    return angle;
};

AngleHint.prototype.getCurrentElementCount = function() {
    var elems = 0;
    for (var i = 0; i < this.list.length; i++) {
      elems += this.getElemCount(this.list[i][1]);
    }
    return elems;
};

AngleHint.prototype.getCombinationAngle = function(sol) {
    var angle = 0;
    for (var i = 0; i < this.list.length; i++) {
      angle += this.list[i][0] * this.getFactor(sol[i]);
    }
    return angle;
};

AngleHint.prototype.getElemCount = function(listEntry) {
    return (listEntry + 1) >> 1;
};

AngleHint.prototype.getFactor = function(listEntry) {
    var cnt = (listEntry + 1) >> 1;
    if ((listEntry & 1) == 0) return cnt;
    return -cnt;
};

AngleHint.prototype.getCombinationElementCount = function(sol) {
    var elems = 0;
    for (var i = 0; i < this.list.length; i++) {
      elems += sol[i] >> 1;
    }
    return elems;
};

AngleHint.prototype.getSolutionArray = function() {
    var sol = new Array(this.list.length);
    for (var i = 0; i < this.list.length; i++) {
      sol[i] = this.list[i][1];
    }
    return sol;
};

AngleHint.prototype.removeDuplicates = function() {
    var ind = 0;
    while (ind < this.solutions.length - 1) {
      var cmp = ind + 1;
      while (cmp < this.solutions.length) {
        if (Math.abs(this.solutions[ind][0][0] - this.solutions[cmp][0][0]) < this.eps) { //(nearly) same angle
          if (this.solutions[ind][0][1] < this.solutions[cmp][0][1]) { //[ind] has less tracks
            this.solutions.splice(cmp, 1);
          }
          else if (this.solutions[ind][0][1] > this.solutions[cmp][0][1]) { //[cmp] has less tracks
            this.solutions.splice(ind, 1);
            ind--; //one step back to have ind++ neutralized
            break;
          }
          else { //do not remove, same angle and same element count => let the user decide
            cmp++;
          }
        }
        else {
          cmp++;
        }
      }
      ind++;
    }
};

function getNormalizedAngle(a, limit, epsilon) {
    a = (Math.round(a / epsilon) % Math.round(limit / epsilon)) * epsilon;
    if (a < 0) {
      a += limit;
    }
    return a;
}

AngleHint.prototype.getAuxPreview = function() {
    var ret = undefined;
    if (isLineEntity(this.lineStart)) {
      ret = new Array();
      addAuxPreviewObjects(this.lineStart, ret);
      if (isLineEntity(this.lineEnd)) {
        addAuxPreviewObjects(this.lineEnd, ret);
        ret.push(new RLine(this.middleStart, this.middleEnd));
      }
    }
    return ret;
};

AngleHint.init = function(basePath) {
    var action = new RGuiAction(QCTrack.MenuItem.AngleHint, RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setScriptFile(basePath + "/AngleHint.js");
    action.setDefaultShortcut(new QKeySequence("q,h"));
    action.setGroupSortOrder(QCTrack.MenuOrder.GroupModification);
    action.setSortOrder(QCTrack.MenuOrder.AngleHint);
    EAction.addGuiActionTo(action, QCTrack, true, false, false);
};

